"""Workflow for creating a new Layer 2 Circuit."""

from typing import Any
from uuid import uuid4

from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from products.product_blocks.service_binding_port import ServiceBindingPortInactive
from products.product_types.layer_2_circuit import Layer2CircuitInactive
from pydantic import BaseModel, ConfigDict, Field
from pydantic_forms.types import FormGenerator, State, UUIDstr
from pydantic_forms.validators import Divider, Label, ReadOnlyField

from gso.products.product_blocks.layer_2_circuit import Layer2CircuitSideBlockInactive, Layer2CircuitType
from gso.utils.helpers import active_edge_port_selector, partner_choice
from gso.utils.types.interfaces import BandwidthString
from gso.utils.types.ip_address import IPv4AddressType, IPV4Netmask, IPv6AddressType, IPV6Netmask
from gso.utils.types.tt_number import TTNumber
from gso.utils.types.virtual_identifiers import VC_ID, VLAN_ID


def initial_input_generator(product_name: str) -> FormGenerator:
    """Gather input from the operator about a new Layer 2 Circuit subscription."""

    class CreateLayer2CircuitServicePage(FormPage):
        model_config = ConfigDict(title=f"{product_name}")

        tt_number: TTNumber
        partner: partner_choice()  # type: ignore[valid-type]
        divider: Divider = Field(None, exclude=True)

        layer_2_circuit_type: Layer2CircuitType
        policer_enabled: bool = False

    initial_user_input = yield CreateLayer2CircuitServicePage

    class Layer2CircuitSideSelection(BaseModel):
        edge_port: active_edge_port_selector(partner_id=initial_user_input.partner)  # type: ignore[valid-type]
        divider: Divider = Field(None, exclude=True)

        geant_sid: str
        is_tagged: bool = False
        vlan_id: VLAN_ID
        ipv4_address: IPv4AddressType
        ipv4_mask: IPV4Netmask
        ipv6_address: IPv6AddressType
        ipv6_mask: IPV6Netmask
        custom_firewall_filters: bool = False

    class Layer2CircuitServiceSidesPage(FormPage):
        model_config = ConfigDict(title=f"{product_name} - Configure Edge Ports")

        if initial_user_input.layer_2_circuit_type == Layer2CircuitType.TAGGED:
            vlan_range_label: Label = Field("Please set a VLAN range, bounds including.", exclude=True)
            vlan_range_lower_bound: VLAN_ID
            vlan_range_upper_bound: VLAN_ID
        else:
            vlan_range_lower_bound: ReadOnlyField(int) = None
            vlan_range_upper_bound: ReadOnlyField(int) = None

        vlan_divider: Divider = Field(None, exclude=True)

        if initial_user_input.policer_enabled:
            policer_bandwidth: BandwidthString
        else:
            policer_bandwidth: ReadOnlyField(str) = None

        policer_divider: Divider = Field(None, exclude=True)

        layer_2_circuit_side_a: Layer2CircuitSideSelection
        side_divider: Divider = Field(None, exclude=True)
        layer_2_circuit_side_b: Layer2CircuitSideSelection

    layer_2_circuit_input = yield Layer2CircuitServiceSidesPage

    return {"product_name": product_name} | initial_user_input.model_dump() | layer_2_circuit_input.model_dump()


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: str) -> State:
    """Create a new subscription object in the database."""
    subscription = Layer2CircuitInactive.from_product_id(product, partner)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: Layer2CircuitInactive,
    layer_2_circuit_side_a: dict[str, Any],
    layer_2_circuit_side_b: dict[str, Any],
    virtual_circuit_id: VC_ID | None,
    layer_2_circuit_type: Layer2CircuitType,
    vlan_range_lower_bound: VLAN_ID | None,
    vlan_range_upper_bound: VLAN_ID | None,
    policer_enabled: bool,  # noqa: FBT001
    policer_bandwidth: BandwidthString | None,
) -> State:
    """Build a subscription object from all user input."""
    subscription.layer_2_circuit.layer_2_circuit_sides = [
        Layer2CircuitSideBlockInactive.new(uuid4(), sbp=ServiceBindingPortInactive.new(uuid4(), **circuit_side))
        for circuit_side in [layer_2_circuit_side_a, layer_2_circuit_side_b]
    ]
    subscription.layer_2_circuit.virtual_circuit_id = virtual_circuit_id
    subscription.layer_2_circuit.layer_2_circuit_type = layer_2_circuit_type
    subscription.layer_2_circuit.vlan_range_lower_bound = vlan_range_lower_bound
    subscription.layer_2_circuit.vlan_range_upper_bound = vlan_range_upper_bound
    subscription.layer_2_circuit.policer_enabled = policer_enabled
    subscription.layer_2_circuit.bandwidth = policer_bandwidth

    return {"subscription": subscription}


@workflow(
    "Create Layer 2 Circuit Service",
    initial_input_form=wrap_create_initial_input_form(initial_input_generator),
    target=Target.CREATE,
)
def create_layer_2_circuit() -> StepList:
    """Create a new Layer 2 Circuit service subscription."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.PROVISIONING)
        >> resync
        >> done
    )
